from hybrid_pool_executor.base import ModuleSpec
from hybrid_pool_executor.workers.process.manager import (
    ProcessManager,
    ProcessManagerSpec,
)
from hybrid_pool_executor.workers.process.worker import (
    ProcessTask,
    ProcessWorker,
    ProcessWorkerSpec,
)

MODULE_SPEC = ModuleSpec(
    name="process",
    manager_type=ProcessManager,
    manager_spec_type=ProcessManagerSpec,
    worker_type=ProcessWorker,
    worker_spec_type=ProcessWorkerSpec,
    tags=frozenset({"process", "thread", "async"}),
    enabled=True,
)


__all__ = (
    "MODULE_SPEC",
    "ProcessManager",
    "ProcessManagerSpec",
    "ProcessTask",
    "ProcessWorker",
    "ProcessWorkerSpec",
)



import dataclasses
import typing as t
from dataclasses import dataclass, field
from queue import Empty, SimpleQueue
from threading import ThreadError
from time import monotonic

from hybrid_pool_executor.base import (
    Action,
    BaseTask,
    BaseWorker,
    BaseWorkerSpec,
    CancelledError,
    Future,
)
from hybrid_pool_executor.constants import (
    ACT_CLOSE,
    ACT_DONE,
    ACT_EXCEPTION,
    ACT_NONE,
    ACT_RESET,
    ACT_RESTART,
    ActionFlag,
    ThreadBus,
    ThreadBusType,
)
from hybrid_pool_executor.utils import (
    AsyncToSync,
    KillableThread,
    WeakClassMethod,
    isasync,
)


@dataclass
class ThreadTask(BaseTask):
    future: Future = field(default_factory=Future)


@dataclass
class ThreadWorkerSpec(BaseWorkerSpec):
    """The specification of thread worker.

    :param task_bus: The queue for sending task item.
    :type task_bus: SimpleQueue

    :param request_bus: The queue for receiving requests from manager.
    :type request_bus: SimpleQueue

    :param response_bus: The queue for sending responses to manager.
    :type response_bus: SimpleQueue

    :param daemon: True if worker should be a daemon, defaults to True.
    :type daemon: bool, optional
    """

    task_bus_type: ThreadBusType = SimpleQueue
    request_bus_type: ThreadBusType = SimpleQueue
    response_bus_type: ThreadBusType = SimpleQueue
    task_bus: t.Optional[ThreadBus] = None
    request_bus: t.Optional[ThreadBus] = None
    response_bus: t.Optional[ThreadBus] = None
    daemon: bool = True


class ThreadWorker(BaseWorker):
    def __init__(self, spec: ThreadWorkerSpec) -> None:
        super().__init__()
        self._spec = dataclasses.replace(spec)
        self._name = self._spec.name
        self._thread: t.Optional[KillableThread] = None
        self._curr_task_name: t.Optional[str] = None

        for checkpoint in ("task_bus", "request_bus", "response_bus"):
            if not getattr(self._spec, checkpoint):
                raise ValueError(f'Param "{checkpoint}" in spec is empty.')

    @property
    def name(self) -> str:
        return self._name

    @property
    def spec(self) -> ThreadWorkerSpec:
        return self._spec

    def _get_response(
        self,
        flag: ActionFlag = ACT_NONE,
        result: t.Optional[t.Any] = None,
        exception: t.Optional[BaseException] = None,
    ) -> Action:
        return Action(
            flag=flag,
            task_name=self._curr_task_name,
            worker_name=self._name,
            result=result,
            exception=exception,
        )

    def start(self):
        if self._state.running or self._thread is not None:
            raise RuntimeError(
                f'{self.__class__.__qualname__} "{self._name}" is already started.'
            )
        self._thread = KillableThread(
            target=WeakClassMethod(self._run),
            daemon=self._spec.daemon,
        )
        self._thread.start()

        # Block method until self._run actually starts to avoid creating multiple
        # workers when in high concurrency situation.
        state = self._state
        while not state.inited:
            pass

    def _run(self):
        state = self._state
        state.running = True
        state.idle = True
        state.inited = True

        spec = self._spec
        get_response = self._get_response
        task_bus: ThreadBus = spec.task_bus
        request_bus: ThreadBus = spec.request_bus
        response_bus: ThreadBus = spec.response_bus
        max_task_count: int = spec.max_task_count
        max_err_count: int = spec.max_err_count
        max_cons_err_count: int = spec.max_cons_err_count
        idle_timeout: float = spec.idle_timeout
        wait_interval: float = spec.wait_interval

        task_count: int = 0
        err_count: int = 0
        cons_err_count: int = 0

        response: t.Optional[Action] = None

        idle_tick = monotonic()
        while True:
            if idle_timeout >= 0 and monotonic() - idle_tick > idle_timeout:
                response = get_response(ACT_CLOSE)
                break
            while not request_bus.empty():
                request: Action = request_bus.get()
                if request.match(ACT_RESET):
                    task_count = 0
                    err_count = 0
                    cons_err_count = 0
                if request.match(ACT_CLOSE, ACT_RESTART):
                    response = get_response(request.flag)
                    break
            if not state.running:
                break

            try:
                task: ThreadTask = task_bus.get(timeout=wait_interval)
            except Empty:
                continue
            result = None
            try:
                state.idle = False
                self._curr_task_name = task.name

                # check if order is cancelled
                if task.cancelled:
                    raise CancelledError(f'Future "{task.name}" has been cancelled')
                if isasync(task.fn):
                    task.fn = t.cast(t.Coroutine[t.Any, t.Any, t.Any], task.fn)
                    sync_coro = AsyncToSync(task.fn, *task.args, **task.kwargs)
                    result = sync_coro()
                else:
                    task.fn = t.cast(t.Callable[..., t.Any], task.fn)
                    result = task.fn(*task.args, **task.kwargs)
            except Exception as exc:
                err_count += 1
                cons_err_count += 1
                task.future.set_exception(exc)
                response = get_response(flag=ACT_EXCEPTION)
            else:
                cons_err_count = 0
                task.future.set_result(result)
                response = get_response(flag=ACT_DONE)
            finally:
                del task
                task_count += 1

                self._curr_task_name = None
                state.idle = True

                idle_tick = monotonic()
                if (
                    0 <= max_task_count <= task_count
                    or 0 <= max_err_count <= err_count
                    or 0 <= max_cons_err_count <= cons_err_count
                ):
                    response = t.cast(Action, response)
                    response.add_flag(ACT_RESTART)
                    break
                response_bus.put(response)
                response = None

        state.idle = False
        state.running = False
        if response is not None and response.flag != ACT_NONE:
            response_bus.put(response)
        self._thread = None

    def stop(self):
        self._state.running = False
        if self._thread and self._thread.is_alive():
            self._thread.join()
        self._thread = None

    def terminate(self):
        self._state.running = False
        try:
            if self._thread and self._thread.is_alive():
                self._thread.terminate()
        except ThreadError:
            pass
        self._thread = None



import asyncio
import sys
import time
import weakref
from random import random

import pytest

from hybrid_pool_executor.base import Action
from hybrid_pool_executor.constants import ACT_EXCEPTION, ACT_RESTART
from hybrid_pool_executor.workers.asyncio import (
    AsyncManager,
    AsyncManagerSpec,
    AsyncTask,
    AsyncWorker,
    AsyncWorkerSpec,
)


@pytest.mark.timeout(10)
def test_async_worker_task():
    async def simple_task():
        return "done"

    worker_spec = AsyncWorkerSpec(name="TestAsyncWorker", idle_timeout=1)
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = AsyncTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)

    worker = AsyncWorker(worker_spec)
    worker.start()

    assert task.future.result() == "done"

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
def test_sync_worker_task():
    def simple_task():
        return "done"

    worker_spec = AsyncWorkerSpec(name="TestAsyncWorker", idle_timeout=1)
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = AsyncTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)

    worker = AsyncWorker(worker_spec)
    worker.start()

    assert task.future.result() == "done"

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_async_worker_task_async_future():
    async def simple_task():
        return "done"

    worker_spec = AsyncWorkerSpec(name="TestAsyncWorker", idle_timeout=1)
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = AsyncTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)

    worker = AsyncWorker(worker_spec)
    worker.start()

    assert await task.future == "done"
    assert task.future.result() == "done"
    assert await task.future == "done"

    worker.stop()


@pytest.mark.timeout(10)
def test_async_worker_error():
    async def simple_error_task():
        raise RuntimeError("error")

    worker_spec = AsyncWorkerSpec(
        name="TestAsyncWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = AsyncTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)

    worker = AsyncWorker(worker_spec)
    worker.start()

    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    response: Action = worker_spec.response_bus.get()
    assert response.match(ACT_EXCEPTION)
    assert response.match(ACT_RESTART)

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task

    # it costs more time to clean up on Windows/MacOS
    if sys.platform != "linux":
        time.sleep(1)

    assert ref() is None


@pytest.mark.timeout(10)
def test_async_worker_max_error():
    async def simple_task():
        return "done"

    async def simple_error_task():
        raise RuntimeError("error")

    worker_spec = AsyncWorkerSpec(
        name="TestAsyncWorker",
        idle_timeout=1,
        max_err_count=2,
        max_cons_err_count=-1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()

    worker = AsyncWorker(worker_spec)
    worker.start()

    task = AsyncTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    assert worker.is_alive()

    task = AsyncTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)
    _ = task.future.result()

    assert worker.is_alive()

    task = AsyncTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    # wait for worker shutdown
    time.sleep(0.25)
    assert not worker.is_alive()


@pytest.mark.timeout(10)
def test_async_worker_cons_error():
    async def simple_error_task():
        raise RuntimeError("error")

    worker_spec = AsyncWorkerSpec(
        name="TestAsyncWorker",
        idle_timeout=1,
        max_err_count=-1,
        max_cons_err_count=2,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()

    worker = AsyncWorker(worker_spec)
    worker.start()

    task = AsyncTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    assert worker.is_alive()

    task = AsyncTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    # wait for worker shutdown
    time.sleep(0.25)
    assert not worker.is_alive()


@pytest.mark.timeout(10)
def test_async_manager():
    async def simple_task():
        return "done"

    manager_spec = AsyncManagerSpec()
    manager = AsyncManager(manager_spec)
    manager.start()

    future = manager.submit(simple_task)
    assert future.result() == "done"

    manager.stop()


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_async_manager_high_concurrency():
    async def simple_task(v):
        await asyncio.sleep(random())
        return v

    with AsyncManager(AsyncManagerSpec()) as manager:
        futures = []
        for i in range(1024):
            futures.append(manager.submit(simple_task, (i,)))
        for i, future in enumerate(futures):
            assert await future == i



import dataclasses
import multiprocessing as mp
import typing as t
from dataclasses import dataclass, field
from functools import partial
from multiprocessing import Process, Value
from multiprocessing.sharedctypes import Synchronized
from queue import Empty
from threading import Event
from time import monotonic

from hybrid_pool_executor.base import (
    Action,
    BaseTask,
    BaseWorker,
    BaseWorkerSpec,
    CancelledError,
    Future,
)
from hybrid_pool_executor.constants import (
    ACT_CLOSE,
    ACT_DONE,
    ACT_EXCEPTION,
    ACT_NONE,
    ACT_RESET,
    ACT_RESTART,
    ActionFlag,
    ProcessBus,
    ProcessBusType,
)
from hybrid_pool_executor.utils import AsyncToSync, isasync
from hybrid_pool_executor.workers.process.queue import Queue


@dataclass
class ProcessTask(BaseTask):
    future: t.Optional[Future] = None


@dataclass
class ProcessWorkerSpec(BaseWorkerSpec):
    task_bus_type: ProcessBusType = Queue
    request_bus_type: ProcessBusType = Queue
    response_bus_type: ProcessBusType = Queue
    task_bus: t.Optional[ProcessBus] = None
    request_bus: t.Optional[ProcessBus] = None
    response_bus: t.Optional[ProcessBus] = None
    task_bus_qsize: Synchronized = field(
        default_factory=t.cast(
            t.Callable[[], Synchronized],
            partial(Value, "L", 0),
        )
    )
    daemon: bool = False


class ProcessWorker(BaseWorker):
    def __init__(self, spec: ProcessWorkerSpec):
        self._spec = dataclasses.replace(spec)
        # TODO: change process name in system monitor
        self._name = self._spec.name

        self._inited = mp.Event()
        self._running = mp.Event()
        self._idle = mp.Event()

        self._process: t.Optional[Process] = None
        self._curr_task_name: t.Optional[str] = None

        for checkpoint in ("task_bus", "request_bus", "response_bus"):
            if not getattr(self._spec, checkpoint):
                raise ValueError(f'Param "{checkpoint}" in spec is empty.')

    @property
    def name(self) -> str:
        return self._name

    @property
    def spec(self) -> ProcessWorkerSpec:
        return self._spec

    def is_alive(self) -> bool:
        return self._running.is_set()

    def is_idle(self) -> bool:
        return self._idle.is_set()

    def start(self):
        if self._running.is_set() or self._process is not None:
            raise RuntimeError(
                f'{self.__class__.__qualname__} "{self._name}" is already started.'
            )
        self._process = Process(
            target=self._run,
            name=self._name,
            kwargs={
                "spec": self._spec,
                "inited": self._inited,
                "idle": self._idle,
                "running": self._running,
            },
            daemon=self._spec.daemon,
        )
        self._process.start()

        while not self._running.is_set():
            pass

    @staticmethod
    def _run(
        spec: ProcessWorkerSpec,
        inited: Event,
        idle: Event,
        running: Event,
    ):
        inited.set()
        idle.set()
        running.set()

        worker_name = spec.name
        curr_task_name: t.Optional[str] = None

        def get_response(
            flag: ActionFlag = ACT_NONE,
            result: t.Optional[t.Any] = None,
            exception: t.Optional[BaseException] = None,
        ) -> Action:
            return Action(
                flag=flag,
                task_name=curr_task_name,
                worker_name=worker_name,
                result=result,
                exception=exception,
            )

        task_bus = t.cast(ProcessBus, spec.task_bus)
        task_bus_qsize = spec.task_bus_qsize
        request_bus = t.cast(ProcessBus, spec.request_bus)
        response_bus = t.cast(ProcessBus, spec.response_bus)
        max_task_count = spec.max_task_count
        max_err_count = spec.max_err_count
        max_cons_err_count = spec.max_cons_err_count
        idle_timeout = spec.idle_timeout
        wait_interval = spec.wait_interval

        task_count = 0
        err_count = 0
        cons_err_count = 0

        response: t.Optional[Action] = None
        should_exit: bool = False
        idle_tick = monotonic()
        while True:
            if idle_timeout >= 0 and monotonic() - idle_tick > idle_timeout:
                response = get_response(ACT_CLOSE)
                break
            while not request_bus.empty():
                request: Action = request_bus.get()
                if request.match(ACT_RESET):
                    task_count = 0
                    err_count = 0
                    cons_err_count = 0
                if request.match(ACT_CLOSE, ACT_RESTART):
                    response = get_response(request.flag)
                    should_exit = True
                    break
            if should_exit or not running.is_set():
                break
            try:
                task: ProcessTask = task_bus.get(timeout=wait_interval)
                task_bus_qsize.value -= 1
            except Empty:
                continue
            result = None
            try:
                idle.clear()
                curr_task_name = task.name

                # check if future is cancelled
                if task.cancelled:
                    raise CancelledError(f'Future "{task.name}" has been cancelled')
                if isasync(task.fn):
                    task.fn = t.cast(t.Coroutine[t.Any, t.Any, t.Any], task.fn)
                    sync_coro = AsyncToSync(task.fn, *task.args, **task.kwargs)
                    result = sync_coro()
                else:
                    task.fn = t.cast(t.Callable[..., t.Any], task.fn)
                    result = task.fn(*task.args, **task.kwargs)
            except Exception as exc:
                err_count += 1
                cons_err_count += 1
                response = get_response(flag=ACT_EXCEPTION, exception=exc)
            else:
                cons_err_count = 0
                response = get_response(flag=ACT_DONE, result=result)
            finally:
                del task
                task_count += 1

                curr_task_name = None
                idle.set()

                idle_tick = monotonic()
                if (
                    0 <= max_task_count <= task_count
                    or 0 <= max_err_count <= err_count
                    or 0 <= max_cons_err_count <= cons_err_count
                ):
                    response = t.cast(Action, response)
                    response.add_flag(ACT_RESTART)
                    break
                response_bus.put(response)
                response = None

        idle.clear()
        running.clear()
        if response is not None and response.flag != ACT_NONE:
            response_bus.put(response)

    def stop(self):
        self._running.clear()
        if self._process and self._process.is_alive():
            self._process.join()
        self._process = None

    def terminate(self):
        self._running.clear()
        if self._process and self._process.is_alive():
            self._process.terminate()
        self._process = None



import importlib
import typing as t

from hybrid_pool_executor.base import ExistsError, ModuleSpec
from hybrid_pool_executor.utils import SingletonMeta

_default_modules = [
    "hybrid_pool_executor.workers.asyncio",
    "hybrid_pool_executor.workers.process",
    "hybrid_pool_executor.workers.thread",
]


class ModuleSpecRepo(dict):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._tag_index: t.Dict[str, t.Set[str]] = {}

    def import_spec(self, spec: ModuleSpec, overwrite: bool = False):
        name = spec.name
        if name in self and not overwrite:
            raise ExistsError(
                f'Spec "{name}" already exists, set overwrite=True if you want to '
                "overwrite existing spec."
            )
        super().__setitem__(name, spec)
        for tag in spec.tags:
            if tag not in self._tag_index:
                self._tag_index[tag] = set()
            index = self._tag_index[tag]
            index.add(name)

    def __getitem__(self, name: str) -> ModuleSpec:
        return t.cast(ModuleSpec, super().__getitem__(name))

    def __setitem__(self, name: str, spec: ModuleSpec) -> None:
        if name != spec.name:
            raise ValueError(
                f'Key ("{name}") and spec name ("{spec.name}") does not match.'
            )
        return self.import_spec(spec=spec, overwrite=True)

    def filter_by_tags(self, *tags: str) -> t.Optional[t.FrozenSet[str]]:
        if not tags:
            return frozenset()
        filter: t.Set[str] = set()
        for tag in tags:
            index = self._tag_index.get(tag)
            if not index:
                return frozenset()
            if not filter:
                filter |= index
            else:
                filter &= index
        return frozenset(filter)


class ModuleSpecFactory(ModuleSpecRepo, metaclass=SingletonMeta):
    def __init__(self):
        super().__init__()
        # load default module specs
        self._import_default()

    def _import_default(self):
        for module in _default_modules:
            self.import_module(module)

    def import_module(self, module: str):
        module_spec: ModuleSpec = t.cast(
            ModuleSpec, importlib.import_module(module).MODULE_SPEC
        )
        self.import_spec(module_spec)

    def get_repo(self) -> ModuleSpecRepo:
        repo = ModuleSpecRepo()
        for spec in self.values():
            repo.import_spec(spec)
        return repo


spec_factory = ModuleSpecFactory()



import asyncio
import typing as t
from dataclasses import dataclass
from queue import Empty, SimpleQueue
from time import monotonic

from hybrid_pool_executor.base import Action, CancelledError
from hybrid_pool_executor.constants import (
    ACT_CLOSE,
    ACT_DONE,
    ACT_EXCEPTION,
    ACT_NONE,
    ACT_RESET,
    ACT_RESTART,
)
from hybrid_pool_executor.utils import iscoroutine, iscoroutinefunction
from hybrid_pool_executor.workers.thread import (
    ThreadTask,
    ThreadWorker,
    ThreadWorkerSpec,
)

NoneType = type(None)


@dataclass
class AsyncTask(ThreadTask):
    ...


@dataclass
class AsyncWorkerSpec(ThreadWorkerSpec):
    max_task_count: int = -1
    max_err_count: int = 10


class AsyncWorker(ThreadWorker):
    def __init__(self, spec: AsyncWorkerSpec):
        super().__init__(spec=t.cast(ThreadWorkerSpec, spec))
        self._spec = t.cast(AsyncWorkerSpec, self._spec)

        self._loop: t.Optional[asyncio.AbstractEventLoop] = None
        self._async_tasks: t.Dict[str, asyncio.Task] = {}
        self._curr_tasks: t.Dict[str, AsyncTask] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def spec(self) -> AsyncWorkerSpec:
        return t.cast(AsyncWorkerSpec, self._spec)

    async def _async_run(self):
        state = self._state
        state.running = True
        state.idle = True
        state.inited = True

        spec = self._spec
        worker_name: str = self._name
        task_bus: SimpleQueue = spec.task_bus
        request_bus: SimpleQueue = spec.request_bus
        response_bus: SimpleQueue = spec.response_bus
        max_task_count: int = spec.max_task_count
        max_err_count: int = spec.max_err_count
        max_cons_err_count: int = spec.max_cons_err_count
        idle_timeout: float = spec.idle_timeout
        wait_interval: float = spec.wait_interval

        loop = t.cast(asyncio.BaseEventLoop, self._loop)
        async_tasks = self._async_tasks
        async_response_bus = asyncio.Queue()

        async def coroutine(
            task: AsyncTask,
            worker_name: str,
            bus: asyncio.Queue,
        ):
            result = resp = None
            try:
                if iscoroutine(task.fn):
                    task.fn = t.cast(t.Coroutine[t.Any, t.Any, t.Any], task.fn)
                    result = await task.fn
                else:
                    task.fn = t.cast(t.Callable[..., t.Any], task.fn)
                    if iscoroutinefunction(task.fn):
                        result = await task.fn(*task.args, **task.kwargs)
                    else:
                        result = task.fn(*task.args, **task.kwargs)
            except Exception as exc:
                resp = Action(
                    flag=ACT_EXCEPTION,
                    task_name=task.name,
                    worker_name=worker_name,
                )
                task.future.set_exception(exc)
                del task.fn
            else:
                resp = Action(
                    flag=ACT_DONE,
                    task_name=task.name,
                    worker_name=worker_name,
                )
                task.future.set_result(result)
            finally:
                await bus.put(resp)

        task_count: int = 0
        err_count: int = 0
        cons_err_count: int = 0
        curr_coroutines: int = 0
        is_prev_coro_err: bool = False

        response = None
        idle_tick = monotonic()
        while True:
            if curr_coroutines > 0:
                state.idle = False
                idle_tick = monotonic()
            else:
                state.idle = True
                if idle_timeout >= 0 and monotonic() - idle_tick > idle_timeout:
                    response = Action(flag=ACT_CLOSE, worker_name=worker_name)
                    break
            while not request_bus.empty():
                request: Action = request_bus.get()
                if request.match(ACT_RESET):
                    task_count = 0
                    err_count = 0
                    cons_err_count = 0
                if request.match(ACT_CLOSE, ACT_RESTART):
                    response = Action(flag=request.flag, worker_name=worker_name)
                    break
            if not state.running:
                break
            try:
                while not 0 <= max_task_count <= task_count:
                    task: AsyncTask = task_bus.get(timeout=wait_interval)
                    # check if future is cancelled
                    if task.cancelled:
                        task_bus.put(
                            Action(
                                flag=ACT_EXCEPTION,
                                task_name=task.name,
                                worker_name=worker_name,
                                exception=CancelledError(
                                    f'Future "{task.name}" has been cancelled'
                                ),
                            )
                        )
                        del task
                        continue
                    else:
                        state.idle = False
                        async_task: asyncio.Task = loop.create_task(
                            coroutine(
                                task=task,
                                worker_name=worker_name,
                                bus=async_response_bus,
                            )
                        )
                        async_tasks[task.name] = async_task
                        del task
                    task_count += 1
                    curr_coroutines += 1
            except Empty:
                pass
            await asyncio.sleep(0)  # ugly but works
            while not async_response_bus.empty():
                response = await async_response_bus.get()
                await async_tasks.pop(response.task_name)
                curr_coroutines -= 1
                if response.match(ACT_EXCEPTION):
                    err_count += 1
                    if is_prev_coro_err:
                        cons_err_count += 1
                    else:
                        cons_err_count = 1
                    is_prev_coro_err = True
                else:
                    cons_err_count = 0
                    is_prev_coro_err = False
                if (
                    0 <= max_task_count <= task_count
                    or 0 <= max_err_count <= err_count
                    or 0 <= max_cons_err_count <= cons_err_count
                ):
                    response.add_flag(ACT_RESTART)
                    response_bus.put(response)
                    state.running = False
                    break
                response_bus.put(response)
                response = None
            if not state.running:
                break
        state.running = False
        for async_task in async_tasks.values():
            if not async_task.done():
                await async_task
        if response is not None and response.flag != ACT_NONE:
            if not response.match(ACT_CLOSE, ACT_RESTART):
                response_bus.put(response)
            else:
                await async_response_bus.put(response)
        while not async_response_bus.empty():
            response = await async_response_bus.get()
            response_bus.put(response)
        self._thread = None

    def _run(self):
        self._loop = asyncio.new_event_loop()
        self._loop.run_until_complete(self._async_run())



import typing as t
from abc import ABC, abstractmethod
from concurrent.futures._base import CancelledError as BaseCancelledError
from concurrent.futures._base import Executor
from concurrent.futures._base import Future as BaseFuture
from dataclasses import dataclass, field

from hybrid_pool_executor.constants import ACT_NONE, ActionFlag, Function
from hybrid_pool_executor.utils import get_event_loop

"""
For python 3.7+, there is no significant speed/size difference between object,
dataclass and namedtuple.
"""

# TODO: migrate common codes into base.py
# TODO: discard exception traceback related to workers


@dataclass
class Action:
    """The base dataclass of action.

    Actions are objects used to transfer information between worker(s) and manager(s)
    through request/response queue.
    """

    flag: ActionFlag = ACT_NONE
    message: t.Optional[str] = None
    task_name: t.Optional[str] = None
    worker_name: t.Optional[str] = None
    result: t.Any = None
    exception: t.Optional[BaseException] = None

    def add_flag(self, flag: ActionFlag):
        self.flag |= flag

    def match(
        self,
        *flags: t.Union[t.Iterable[ActionFlag], ActionFlag],
        strategy: t.Literal["all", "any"] = "any",
    ) -> bool:
        if strategy not in ("any", "all"):
            raise ValueError(
                'Param "strategy" should be "any" or "all", '
                f'got "{strategy}" instread".'
            )
        if len(flags) == 1 and isinstance(flags[0], (tuple, list, set)):
            flags = tuple(flags[0])
        flags = t.cast(t.Tuple[ActionFlag, ...], flags)
        m = map(lambda flag: self.flag & flag, flags)
        return any(m) if strategy == "any" else all(m)


@dataclass
class WorkerState:
    """A dataclass which stores running state of worker.

    A bare bool flag may not be synced in concurrent situation, this is why we need
    a WorkerState class.
    """

    inited: bool = False
    running: bool = False
    idle: bool = False


@dataclass
class BaseTask(ABC):
    """The base dataclass of task.

    Calls are objects used to carry functions to worker(s).

    BaseCall is regarded as a abstract class and should not be initialized directly.
    """

    name: str
    fn: Function
    args: t.Iterable[t.Any] = ()
    kwargs: t.Dict[str, t.Any] = field(default_factory=dict)
    cancelled: bool = False


@dataclass
class BaseWorkerSpec(ABC):
    """The base dataclass of worker specification.

    BaseWorkerSpec is regarded as a abstract class and should not be initialized
    directly.

    :param name: Name of worker.
    :type name: str

    :param idle_timeout: Second(s) before the worker should exit after being idle,
        defaults to 60.
    :type idle_timeout: float, optional

    :param wait_interval: Interval in second(s) the worker fetches information,
        defaults to 0.1.
    :type wait_interval: float, optional

    :param max_task_count: Maximum task amount the worker can process, after that the
        worker should be destroyed, defaults to 12, negative value means unlimited.
    :type max_task_count: int, optional

    :param max_err_count: Maximum error amount the worker can afford, after that the
        worker should be destroyed, defaults to 3, negative value means unlimited.
    :type max_err_count: int, optional

    :param max_cons_err_count: Maximum continuous error amount the worker can afford,
        after that the worker should be destroyed, defaults to -1, negative value means
        unlimited.
    :type max_cons_err_count: int, optional
    """

    name: str
    idle_timeout: float = 60.0
    wait_interval: float = 0.1
    max_task_count: int = 12
    max_err_count: int = 3
    max_cons_err_count: int = -1


class BaseWorker(ABC):
    def __init__(self, *args, **kwargs):
        self._state = WorkerState()

    def is_alive(self) -> bool:
        return self._state.running

    def is_idle(self) -> bool:
        return self._state.idle if self._state.running else False

    def _ensure_running(self) -> None:
        if not self._state.running:
            raise RuntimeError(
                f"{self.__class__.__name__} is either stopped or not started "
                "yet and not able to accept tasks."
            )

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self, timeout: t.Optional[float] = None):
        ...

    @abstractmethod
    def terminate(self):
        ...

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class Future(BaseFuture):
    def __init__(self):
        super().__init__()
        self._loop = get_event_loop()
        self._async_fut = self._loop.create_future()

        def cb(_):
            self._async_fut.get_loop().call_soon_threadsafe(
                self._async_fut.set_result, True
            )

        self.add_done_callback(cb)

    async def _async_result(self):
        await self._async_fut
        return self.result()

    def __await__(self):
        return self._async_result().__await__()


BaseExecutor = Executor
CancelledError = BaseCancelledError


class ExistsError(Exception):
    ...


class NotSupportedError(Exception):
    ...


@dataclass
class BaseManagerSpec(ABC):
    name_pattern: str = "BaseManager-{manager_seq}"
    worker_name_pattern: str = "BaseWorker-{worker_seq} [{manager}]"
    task_name_pattern: str = "BaseTask-{task_seq} [{manager}]"
    num_workers: int = -1
    incremental: bool = True
    wait_interval: float = 0.1
    idle_timeout: float = 60.0


class BaseManager(BaseWorker):
    @t.overload
    def submit(
        self,
        fn: t.Coroutine[t.Any, t.Any, t.Any],
        *,
        name: t.Optional[str] = None,
    ) -> Future:
        ...

    @t.overload
    def submit(
        self,
        fn: t.Callable[..., t.Any],
        args: t.Optional[t.Iterable[t.Any]] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
    ) -> Future:
        ...

    @abstractmethod
    def submit(
        self,
        fn: Function,
        args: t.Optional[t.Iterable[t.Any]] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
    ) -> Future:
        ...


def adjust_worker_iterator(
    spec: BaseManagerSpec,
    curr_workers: t.Iterable[BaseWorker],
    num_curr_tasks: int,
) -> range:
    if spec.incremental or spec.num_workers < 0:
        num_idle_workers: int = sum(1 if w.is_idle() else 0 for w in curr_workers)
        if spec.num_workers < 0:
            iterator = range(num_curr_tasks - num_idle_workers)
        else:
            num_curr_workers: int = len(t.cast(t.Sized, curr_workers))
            iterator = range(
                num_curr_workers,
                min(
                    spec.num_workers,
                    num_curr_workers + num_curr_tasks - num_idle_workers,
                ),
            )
    else:
        iterator = range(len(t.cast(t.Sized, curr_workers)), spec.num_workers)
    return iterator


@dataclass
class ModuleSpec:
    name: str
    manager_type: t.Union[
        t.Type[BaseManager],
        t.Callable[[BaseManagerSpec], BaseManager],
    ]
    manager_spec_type: t.Union[
        t.Type[BaseManagerSpec],
        t.Callable[[], BaseManagerSpec],
    ]
    worker_type: t.Union[
        t.Type[BaseWorker],
        t.Callable[[BaseWorkerSpec], BaseWorker],
    ]
    worker_spec_type: t.Union[
        t.Type[BaseWorkerSpec],
        t.Callable[[], BaseWorkerSpec],
    ]
    tags: t.FrozenSet[str]
    enabled: bool = True

    def __post_init__(self):
        assert self.name in self.tags



import dataclasses
import itertools
import os
import typing as t
from dataclasses import dataclass, field
from functools import partial
from multiprocessing import Value
from threading import Event, ThreadError
from time import monotonic

from hybrid_pool_executor.base import (
    Action,
    BaseManager,
    BaseManagerSpec,
    Function,
    Future,
    adjust_worker_iterator,
)
from hybrid_pool_executor.constants import (
    ACT_CLOSE,
    ACT_DONE,
    ACT_EXCEPTION,
    ACT_RESTART,
)
from hybrid_pool_executor.utils import (
    KillableThread,
    WeakClassMethod,
    coalesce,
    rectify,
)
from hybrid_pool_executor.workers.process.worker import (
    ProcessTask,
    ProcessWorker,
    ProcessWorkerSpec,
)


@dataclass
class ProcessManagerSpec(BaseManagerSpec):
    name_pattern: str = "ProcessManager-{manager_seq}"
    worker_name_pattern: str = "ProcessWorker-{worker_seq} [{manager}]"
    task_name_pattern: str = "ProcessTask-{task_seq} [{manager}]"
    # -1: unlimited; 0: same as num_workers
    max_processing_responses_per_iteration: int = -1
    idle_timeout: float = 60.0
    task_class: t.Type[ProcessTask] = ProcessTask
    worker_class: t.Type[ProcessWorker] = ProcessWorker
    worker_spec: ProcessWorkerSpec = field(
        default_factory=partial(ProcessWorkerSpec, name="DefaultWorkerSpec")
    )

    def __post_init__(self):
        if self.num_workers == 0:
            self.num_workers == os.cpu_count() * 2
        if self.max_processing_responses_per_iteration == 0:
            self.max_processing_responses_per_iteration = self.num_workers


class ProcessManager(BaseManager):
    _next_manager_seq = itertools.count().__next__

    def __init__(self, spec: ProcessManagerSpec):
        super().__init__()
        self._spec = dataclasses.replace(spec)
        self._worker_class = self._spec.worker_class
        self._worker_spec = dataclasses.replace(spec.worker_spec)
        self._name = self._spec.name_pattern.format(
            manager_seq=self.__class__._next_manager_seq()
        )

        self._next_worker_seq = itertools.count().__next__
        self._next_task_seq = itertools.count().__next__

        self._task_bus = self._worker_spec.task_bus_type()
        self._task_bus_qsize = Value("L", 0)
        self._response_bus = self._worker_spec.response_bus_type()
        self._curr_workers: t.Dict[str, ProcessWorker] = {}
        self._curr_tasks: t.Dict[str, t.Any] = {}
        self._thread: t.Optional[KillableThread] = None

    def start(self):
        if self._state.running or self._thread is not None:
            raise RuntimeError(f'ThreadManager "{self._name}" is already started.')
        self._thread = KillableThread(
            target=WeakClassMethod(self._run),
            daemon=True,
        )
        self._thread.start()

        state = self._state
        while not state.inited:
            pass

    def _run(self):
        state = self._state
        state.running = True
        state.inited = True

        metronome = Event()
        wait_interval: float = rectify(coalesce(self._spec.wait_interval, 0.1), 0.1)
        idle_timeout: float = rectify(coalesce(self._spec.idle_timeout, 60), 60)
        num_process_limit: int = rectify(
            coalesce(self._spec.max_processing_responses_per_iteration, -1), -1
        )
        curr_tasks = self._curr_tasks
        consume_response = self._consume_response
        response_bus = self._response_bus

        idle_tick = monotonic()
        while True:
            if not curr_tasks and response_bus.empty():
                if idle_timeout >= 0 and monotonic() - idle_tick > idle_timeout:
                    break
            else:
                idle_tick = monotonic()
            if not state.running:
                break

            num_processed: int = 0
            while not response_bus.empty():
                consume_response()
                num_processed += 1
                if num_processed >= num_process_limit:
                    break
            if num_processed == 0:
                metronome.wait(wait_interval)

        state.running = False
        # TODO: this part blocks function to make sure all processes' result be
        #       captured, however this may also block the stop/shutdown procedure,
        #       especially when using "with" statement (blocked on __exit__() until all
        #       result has been sent back). Maybe we need a better implementation.
        while self._curr_tasks:
            consume_response()
        self._stop_all_workers()
        self._thread = None

    def _stop_all_workers(self):
        stop_action = Action(ACT_CLOSE)
        for worker in self._curr_workers.values():
            worker.spec.request_bus.put(stop_action)
        for worker in self._curr_workers.values():
            worker.stop()

    def stop(self, timeout: t.Optional[float] = 60.0):
        self._state.running = False
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout=timeout)
            if self._thread and self._thread.is_alive():
                raise RuntimeError(f"Failed to stop {self.__class__.__name__}")
        self._thread = None

    def terminate(self):
        self._state.running = False
        try:
            if self._thread and self._thread.is_alive():
                self._thread.terminate()
        except ThreadError:
            pass
        self._thread = None

    def get_worker_spec(
        self,
        name: t.Optional[str] = None,
        daemon: t.Optional[bool] = None,
        idle_timeout: t.Optional[float] = None,
        wait_interval: t.Optional[float] = None,
        max_task_count: t.Optional[int] = None,
        max_err_count: t.Optional[int] = None,
        max_cons_err_count: t.Optional[int] = None,
    ) -> ProcessWorkerSpec:
        if name and name in self._curr_tasks:
            raise KeyError(f'Worker "{name}" exists.')
        default_spec = self._worker_spec
        worker_spec = default_spec.__class__(
            name=coalesce(
                name,
                self._spec.worker_name_pattern.format(
                    manager=self._name,
                    worker_seq=self._next_worker_seq(),
                ),
            ),
            task_bus=self._task_bus,
            task_bus_qsize=self._task_bus_qsize,
            request_bus=default_spec.request_bus_type(),
            response_bus=self._response_bus,
            daemon=coalesce(daemon, default_spec.daemon),
            idle_timeout=rectify(
                coalesce(idle_timeout, default_spec.idle_timeout),
                default_spec.idle_timeout,
            ),
            wait_interval=rectify(
                coalesce(wait_interval, default_spec.wait_interval),
                default_spec.wait_interval,
            ),
            max_task_count=rectify(
                coalesce(max_task_count, default_spec.max_task_count),
                default_spec.max_task_count,
            ),
            max_err_count=rectify(
                coalesce(max_err_count, default_spec.max_err_count),
                default_spec.max_err_count,
            ),
            max_cons_err_count=rectify(
                coalesce(max_cons_err_count, default_spec.max_cons_err_count),
                default_spec.max_cons_err_count,
            ),
        )
        return worker_spec

    def _get_task_name(self, name: t.Optional[str] = None) -> str:
        if name:
            if name in self._curr_tasks:
                raise KeyError(f'Task "{name}" exists.')
            return name
        return coalesce(
            name,
            self._spec.task_name_pattern.format(
                manager=self._name,
                task_seq=self._next_task_seq(),
            ),
        )

    def submit(
        self,
        fn: Function,
        args: t.Optional[t.Iterable[t.Any]] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
    ) -> Future:
        self._ensure_running()
        name = self._get_task_name(name)
        future = Future()
        task = self._spec.task_class(
            name=name,
            fn=fn,
            args=args or (),
            kwargs=kwargs or {},
            future=future,
        )
        self._curr_tasks[name] = task
        # exclude future when sending to other process to reduce size
        self._task_bus.put(dataclasses.replace(task, future=None))
        self._task_bus_qsize.value += 1
        self._adjust_workers()
        return future

    def _consume_response(self):
        response: Action = self._response_bus.get()
        response.task_name = t.cast(str, response.task_name)
        response.worker_name = t.cast(str, response.worker_name)
        if response.match(ACT_DONE, ACT_EXCEPTION):
            task: ProcessTask = self._curr_tasks.pop(response.task_name)
            task.future = t.cast(Future, task.future)
            if response.match(ACT_DONE):
                task.future.set_result(response.result)
            else:
                task.future.set_exception(response.exception)
        if response.match(ACT_CLOSE):
            self._curr_workers.pop(response.worker_name)
        elif response.match(ACT_RESTART):
            worker = self._curr_workers[response.worker_name]
            worker.stop()
            worker.start()

    def _adjust_workers(self):
        # return if the number of workers already meets requirements
        # works on both incremental and static mode
        if len(self._curr_workers) == self._spec.num_workers:
            return
        # if more workers are needed, create them
        for _ in adjust_worker_iterator(
            spec=self._spec,
            curr_workers=self._curr_workers.values(),
            num_curr_tasks=self._task_bus_qsize.value,
        ):
            worker = self._worker_class(self.get_worker_spec())
            self._curr_workers[worker._name] = worker
            worker.start()



import asyncio
import sys
import time
import weakref
from random import random

import pytest

try:
    import cloudpickle  # noqa: F401

    skip_cloudpickle_test = False
except ImportError:
    skip_cloudpickle_test = True


from hybrid_pool_executor.base import Action
from hybrid_pool_executor.constants import ACT_EXCEPTION, ACT_RESTART
from hybrid_pool_executor.workers.process import (
    ProcessManager,
    ProcessManagerSpec,
    ProcessTask,
    ProcessWorker,
    ProcessWorkerSpec,
)


def simple_task():
    return "done"


def simple_error_task():
    raise RuntimeError("error")


def simple_task_v(v):
    time.sleep(random())
    return v


async def simple_async_task_v(v):
    await asyncio.sleep(random())
    return v


@pytest.mark.timeout(10)
def test_process_worker_task():
    worker_spec = ProcessWorkerSpec(
        name="TestProcessWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = ProcessTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)

    worker = ProcessWorker(worker_spec)
    worker.start()

    response: Action = worker_spec.response_bus.get()
    assert response.result == "done"

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
def test_processs_worker_async_task():
    worker_spec = ProcessWorkerSpec(
        name="TestThreadWorker",
        max_task_count=3,
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    tasks = []
    for i in range(3):
        task = ProcessTask(name="simple_async_task", fn=simple_async_task_v, args=[i])
        tasks.append(task)
        worker_spec.task_bus.put(task)

    worker = ProcessWorker(worker_spec)
    worker.start()

    for i in range(3):
        response: Action = worker_spec.response_bus.get()
        assert response.result == i

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
def test_process_worker_error():
    worker_spec = ProcessWorkerSpec(
        name="TestProcessWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = ProcessTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)

    worker = ProcessWorker(worker_spec)
    worker.start()

    response: Action = worker_spec.response_bus.get()
    assert isinstance(response.exception, RuntimeError)
    assert response.match(ACT_EXCEPTION)
    assert response.match(ACT_RESTART)

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()


@pytest.mark.timeout(10)
def test_process_manager():
    manager_spec = ProcessManagerSpec()
    manager = ProcessManager(manager_spec)
    manager.start()

    future = manager.submit(simple_task)
    assert future.result() == "done"

    manager.stop()


@pytest.mark.timeout(20 if sys.platform == "linux" else 60)
@pytest.mark.asyncio
async def test_process_manager_high_concurrency():
    with ProcessManager(ProcessManagerSpec()) as manager:
        futures = []
        for i in range(32):
            futures.append(manager.submit(simple_task_v, (i,)))
        for i, future in enumerate(futures):
            assert await future == i


@pytest.mark.skipif(skip_cloudpickle_test, reason="cloudpickle is not installed")
@pytest.mark.timeout(10)
def test_cloudpickle_process_manager():
    def clousure_task():
        return "done"

    async def async_clousure_task():
        return "done"

    lambda_task = lambda x: x  # noqa: E731

    manager_spec = ProcessManagerSpec()
    manager = ProcessManager(manager_spec)
    manager.start()

    future = manager.submit(clousure_task)
    assert future.result() == "done"

    future = manager.submit(async_clousure_task)
    assert future.result() == "done"

    future = manager.submit(lambda_task, args=("done",))
    assert future.result() == "done"

    manager.stop()



import typing as t
from dataclasses import dataclass, field
from functools import partial

from hybrid_pool_executor.workers.asyncio.worker import (
    AsyncTask,
    AsyncWorker,
    AsyncWorkerSpec,
)
from hybrid_pool_executor.workers.thread import (
    ThreadManager,
    ThreadManagerSpec,
    ThreadWorker,
)

NoneType = type(None)


@dataclass
class AsyncManagerSpec(ThreadManagerSpec):
    name_pattern: str = "AsyncManager-{manager_seq}"
    worker_name_pattern: str = "AsyncWorker-{worker_seq} [{manager}]"
    task_name_pattern: str = "AsyncTask-{task_seq} [{manager}]"
    num_workers: int = 1
    task_class: t.Type[AsyncTask] = AsyncTask
    worker_class: t.Type[AsyncWorker] = AsyncWorker
    worker_spec: AsyncWorkerSpec = field(
        default_factory=partial(AsyncWorkerSpec, name="DefaultWorkerSpec")
    )


class AsyncManager(ThreadManager):
    def __init__(self, spec: AsyncManagerSpec):
        super().__init__(spec=t.cast(ThreadManagerSpec, spec))
        self._spec = t.cast(AsyncManagerSpec, self._spec)
        self._worker_spec = t.cast(AsyncWorkerSpec, self._worker_spec)
        self._curr_workers = t.cast(t.Dict[str, ThreadWorker], self._curr_workers)

    def get_worker_spec(
        self,
        name: t.Optional[str] = None,
        daemon: t.Optional[bool] = None,
        idle_timeout: t.Optional[float] = None,
        wait_interval: t.Optional[float] = None,
        max_task_count: t.Optional[int] = None,
        max_err_count: t.Optional[int] = None,
        max_cons_err_count: t.Optional[int] = None,
    ) -> AsyncWorkerSpec:
        return t.cast(
            AsyncWorkerSpec,
            super().get_worker_spec(
                name=name,
                daemon=daemon,
                idle_timeout=idle_timeout,
                wait_interval=wait_interval,
                max_task_count=max_task_count,
                max_err_count=max_err_count,
                max_cons_err_count=max_cons_err_count,
            ),
        )



from hybrid_pool_executor.base import Future
from hybrid_pool_executor.executor import HybridPoolExecutor
from hybrid_pool_executor.spec import spec_factory as SpecFactory

__version__ = "0.0.4"
__author__ = "Leavers"
__contact__ = "leavers930@gmail.com"

__all__ = (
    "Future",
    "HybridPoolExecutor",
    "SpecFactory",
)



import asyncio
import time
import weakref
from random import random

import pytest

from hybrid_pool_executor.base import Action
from hybrid_pool_executor.constants import ACT_EXCEPTION, ACT_RESTART
from hybrid_pool_executor.workers.thread import (
    ThreadManager,
    ThreadManagerSpec,
    ThreadTask,
    ThreadWorker,
    ThreadWorkerSpec,
)


@pytest.mark.timeout(10)
def test_thread_worker_task():
    def simple_task():
        return "done"

    worker_spec = ThreadWorkerSpec(
        name="TestThreadWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = ThreadTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)

    worker = ThreadWorker(worker_spec)
    worker.start()

    assert task.future.result() == "done"

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
def test_thread_worker_async_task():
    async def simple_task(v):
        await asyncio.sleep(0.1)
        return v

    worker_spec = ThreadWorkerSpec(
        name="TestThreadWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    tasks = []
    for i in range(3):
        task = ThreadTask(name="simple_task", fn=simple_task, args=[i])
        tasks.append(task)
        worker_spec.task_bus.put(task)

    worker = ThreadWorker(worker_spec)
    worker.start()

    for i in range(3):
        assert tasks[i].future.result() == i

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()

    ref = weakref.ref(worker)
    del worker_spec, worker, task
    assert ref() is None


@pytest.mark.timeout(10)
def test_thread_worker_error():
    def simple_error_task():
        raise RuntimeError("error")

    worker_spec = ThreadWorkerSpec(
        name="TestThreadWorker",
        idle_timeout=1,
        max_err_count=1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()
    task = ThreadTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)

    worker = ThreadWorker(worker_spec)
    worker.start()

    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    response: Action = worker_spec.response_bus.get()
    assert response.match(ACT_EXCEPTION)
    assert response.match(ACT_RESTART)

    worker.stop()
    assert not worker.is_alive()
    assert not worker.is_idle()


@pytest.mark.timeout(10)
def test_thread_worker_max_error():
    def simple_task():
        return "done"

    def simple_error_task():
        raise RuntimeError("error")

    worker_spec = ThreadWorkerSpec(
        name="TestThreadWorker",
        idle_timeout=1,
        max_err_count=2,
        max_cons_err_count=-1,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()

    worker = ThreadWorker(worker_spec)
    worker.start()

    task = ThreadTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    assert worker.is_alive()

    task = ThreadTask(name="simple_task", fn=simple_task)
    worker_spec.task_bus.put(task)
    _ = task.future.result()

    assert worker.is_alive()

    task = ThreadTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    # wait for worker shutting down
    time.sleep(0.25)

    assert not worker.is_alive()


@pytest.mark.timeout(10)
def test_thread_worker_cons_error():
    def simple_error_task():
        raise RuntimeError("error")

    worker_spec = ThreadWorkerSpec(
        name="TestThreadWorker",
        idle_timeout=1,
        max_err_count=-1,
        max_cons_err_count=2,
    )
    worker_spec.task_bus = worker_spec.task_bus_type()
    worker_spec.request_bus = worker_spec.request_bus_type()
    worker_spec.response_bus = worker_spec.response_bus_type()

    worker = ThreadWorker(worker_spec)
    worker.start()

    task = ThreadTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    assert worker.is_alive()

    task = ThreadTask(name="simple_error_task", fn=simple_error_task)
    worker_spec.task_bus.put(task)
    with pytest.raises(RuntimeError, match="error"):
        _ = task.future.result()

    # wait for worker shutting down
    time.sleep(0.25)

    assert not worker.is_alive()


@pytest.mark.timeout(10)
def test_thread_manager():
    def simple_task():
        return "done"

    manager_spec = ThreadManagerSpec()
    manager = ThreadManager(manager_spec)
    manager.start()

    future = manager.submit(simple_task)
    assert future.result() == "done"

    manager.stop()


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_thread_manager_high_concurrency():
    def simple_task(v):
        time.sleep(random())
        return v

    # NOTE: The more threads are created, the more time they takes to stop, and the
    # longer the pytest timeout tolerance is required.
    with ThreadManager(ThreadManagerSpec()) as manager:
        futures = []
        for i in range(64):
            futures.append(manager.submit(simple_task, (i,)))
        for i, future in enumerate(futures):
            assert await future == i



from hybrid_pool_executor.base import ModuleSpec
from hybrid_pool_executor.workers.asyncio.manager import AsyncManager, AsyncManagerSpec
from hybrid_pool_executor.workers.asyncio.worker import (
    AsyncTask,
    AsyncWorker,
    AsyncWorkerSpec,
)

MODULE_SPEC = ModuleSpec(
    name="async",
    manager_type=AsyncManager,
    manager_spec_type=AsyncManagerSpec,
    worker_type=AsyncWorker,
    worker_spec_type=AsyncWorkerSpec,
    tags=frozenset({"async"}),
    enabled=True,
)


__all__ = (
    "MODULE_SPEC",
    "AsyncManager",
    "AsyncManagerSpec",
    "AsyncTask",
    "AsyncWorker",
    "AsyncWorkerSpec",
)



from hybrid_pool_executor.base import ModuleSpec
from hybrid_pool_executor.workers.thread.manager import ThreadManager, ThreadManagerSpec
from hybrid_pool_executor.workers.thread.worker import (
    ThreadTask,
    ThreadWorker,
    ThreadWorkerSpec,
)

MODULE_SPEC = ModuleSpec(
    name="thread",
    manager_type=ThreadManager,
    manager_spec_type=ThreadManagerSpec,
    worker_type=ThreadWorker,
    worker_spec_type=ThreadWorkerSpec,
    tags=frozenset({"thread", "async"}),
    enabled=True,
)


__all__ = (
    "MODULE_SPEC",
    "ThreadManager",
    "ThreadManagerSpec",
    "ThreadTask",
    "ThreadWorker",
    "ThreadWorkerSpec",
)



import asyncio
import ctypes
import functools
import inspect
import typing as t
import weakref
from operator import le
from threading import Lock, Thread
from types import MethodType

T = t.TypeVar("T")


def coalesce(*args) -> t.Any:
    for arg in args:
        if arg is not None:
            return arg
    return None


@functools.singledispatch
def rectify(val, fallback=None, threshold=None, operator=le):
    return fallback if operator(val, threshold) else val


@rectify.register(int)
def _(
    val: int,
    fallback: int = -1,
    threshold: int = 0,
    operator: t.Callable = le,
) -> int:
    return fallback if operator(val, threshold) else val


@rectify.register(float)
def _(
    val: float,
    fallback: float = -1.0,
    threshold: float = 0.0,
    operator: t.Callable = le,
) -> float:
    return fallback if operator(val, threshold) else val


@rectify.register(type(None))  # type: ignore
def _(*args, **kwargs):
    raise TypeError('Param "val" should not be None.')


iscoroutine = inspect.iscoroutine
iscoroutinefunction = inspect.iscoroutinefunction
ismethod = inspect.ismethod


def isasync(object: t.Any):
    return iscoroutine(object) or iscoroutinefunction(object)


_singleton_instances = {}
_singleton_lock = Lock()


class Singleton:
    def __new__(cls, *args, **kwargs):
        if cls not in _singleton_instances:
            with _singleton_lock:
                if cls not in _singleton_instances:
                    _singleton_instances[cls] = object.__new__(cls, *args, **kwargs)
        return _singleton_instances[cls]


class SingletonMeta(type):
    def __call__(cls, *args, **kwargs):
        if cls not in _singleton_instances:
            with _singleton_lock:
                if cls not in _singleton_instances:
                    _singleton_instances[cls] = super().__call__(*args, **kwargs)
        return _singleton_instances[cls]


def get_event_loop(
    loop: t.Optional[asyncio.AbstractEventLoop] = None,
) -> asyncio.AbstractEventLoop:
    if loop:
        return loop
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


class AsyncToSync:
    def __init__(self, fn, /, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.is_coro = iscoroutine(self.fn)
        self.is_func = iscoroutinefunction(self.fn)

    def __call__(self, loop=None):
        if not self.is_coro and not self.is_func:
            return self.fn(*self.args, **self.kwargs)
        if self.is_func:
            self.fn = self.fn(*self.args, **self.kwargs)
        loop = get_event_loop(loop)
        if loop.is_running():
            raise RuntimeError("Unable to execute when loop is already running.")
        return loop.run_until_complete(self.fn)


class KillableThread(Thread):
    @staticmethod
    def _raise_to_kill(tid, exctype):
        if not inspect.isclass(exctype):
            raise TypeError("Only types can be raised (not instances)")
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        if res == 0:
            raise ValueError(f'Invalid thread id "{tid}"')
        elif res != 1:
            # if it returns a number greater than one, you're in trouble,
            # and you should call it again with exc=NULL to revert the effect
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
            raise SystemError("PyThreadState_SetAsyncExc failed.")

    def raise_exc(self, exctype):
        self._raise_to_kill(self.ident, exctype)

    def terminate(self):
        self.raise_exc(SystemExit)


class WeakClassMethod:
    def __init__(self, method: MethodType) -> None:
        if not ismethod(method):
            raise TypeError(f"Object {method} is not a class method")
        self._cls_ref = weakref.ref(method.__self__)
        self._name = method.__name__

    def __call__(self, *args, **kwargs) -> t.Any:
        return getattr(self._cls_ref(), self._name)(*args, **kwargs)



import collections
import errno
import os
import sys
import threading
import time
import typing as t
import weakref
from multiprocessing import Value, connection
from multiprocessing import context as _ctx
from multiprocessing import get_context
from multiprocessing.queues import Queue as BaseQueue
from multiprocessing.reduction import ForkingPickler
from multiprocessing.util import Finalize, debug, info, is_exiting, register_after_fork
from queue import Empty, Full

try:
    import cloudpickle

    _ForkingPickler = cloudpickle
except ImportError:
    _ForkingPickler = ForkingPickler


_sentinel = object()


class Queue(BaseQueue):
    def __init__(self, maxsize: int = 0, *, ctx: t.Optional[_ctx.BaseContext] = None):
        if maxsize <= 0:
            # Can raise ImportError (see issues #3770 and #23400)
            from multiprocessing.synchronize import SEM_VALUE_MAX  # type: ignore

            maxsize = SEM_VALUE_MAX

        if not ctx:
            ctx = get_context()

        self._maxsize = maxsize
        self._reader, self._writer = connection.Pipe(duplex=False)
        self._rlock = ctx.Lock()
        self._opid = os.getpid()
        self._qsize = Value("L", 0)  # type: ignore
        if sys.platform == "win32":
            self._wlock = None
        else:
            self._wlock = ctx.Lock()
        self._sem = ctx.BoundedSemaphore(maxsize)
        # For use by concurrent.futures
        self._ignore_epipe = False
        self._reset()

        if sys.platform != "win32":
            register_after_fork(self, Queue._after_fork)

    def __getstate__(self):
        _ctx.assert_spawning(self)
        return (
            self._ignore_epipe,
            self._maxsize,
            self._reader,
            self._writer,
            self._rlock,
            self._wlock,
            self._sem,
            self._opid,
            self._qsize,
        )

    def __setstate__(self, state):
        (
            self._ignore_epipe,
            self._maxsize,
            self._reader,
            self._writer,
            self._rlock,
            self._wlock,
            self._sem,
            self._opid,
            self._qsize,
        ) = state
        self._reset()

    def _after_fork(self):
        debug("Queue._after_fork()")
        self._reset(after_fork=True)

    def _reset(self, after_fork=False):
        if after_fork:
            self._notempty._at_fork_reinit()
        else:
            self._notempty = threading.Condition(threading.Lock())
        self._buffer = collections.deque()
        self._thread = None
        self._jointhread = None
        self._joincancelled = False
        self._closed = False
        self._close = None
        self._send_bytes = self._writer.send_bytes
        self._recv_bytes = self._reader.recv_bytes
        self._poll = self._reader.poll

    def put(self, obj, block=True, timeout=None):
        if self._closed:
            raise ValueError(f"Queue {self!r} is closed")
        if not self._sem.acquire(block, timeout):
            raise Full
        self._qsize.value += 1

        with self._notempty:
            if self._thread is None:
                self._start_thread()
            self._buffer.append(obj)
            self._notempty.notify()

    def get(self, block=True, timeout=None):
        if self._closed:
            raise ValueError(f"Queue {self!r} is closed")
        if block and timeout is None:
            with self._rlock:
                res = self._recv_bytes()
            self._sem.release()
        else:
            if block:
                deadline = time.monotonic() + timeout
            if not self._rlock.acquire(block, timeout):
                raise Empty
            try:
                if block:
                    timeout = deadline - time.monotonic()
                    if not self._poll(timeout):
                        raise Empty
                elif not self._poll():
                    raise Empty
                res = self._recv_bytes()
                self._sem.release()
            finally:
                self._rlock.release()
        # unserialize the data after having released the lock
        result = _ForkingPickler.loads(res)
        self._qsize.value -= 1
        return result

    def qsize(self):
        # Raises NotImplementedError on Mac OSX because of broken sem_getvalue()
        # return self._maxsize - self._sem._semlock._get_value()
        return self._qsize.value

    def _start_thread(self):
        debug("Queue._start_thread()")

        # Start thread which transfers data from buffer to pipe
        self._buffer.clear()
        self._thread = threading.Thread(
            target=Queue._feed,
            args=(
                self._buffer,
                self._notempty,
                self._send_bytes,
                self._wlock,
                self._writer.close,
                self._ignore_epipe,
                self._on_queue_feeder_error,
                self._sem,
                self._qsize,
            ),
            name="QueueFeederThread",
        )
        self._thread.daemon = True

        debug("doing self._thread.start()")
        self._thread.start()
        debug("... done self._thread.start()")

        if not self._joincancelled:
            self._jointhread = Finalize(
                self._thread,
                Queue._finalize_join,
                [weakref.ref(self._thread)],
                exitpriority=-5,
            )

        # Send sentinel to the thread queue object when garbage collected
        self._close = Finalize(
            self, Queue._finalize_close, [self._buffer, self._notempty], exitpriority=10
        )

    @staticmethod
    def _finalize_join(twr):
        debug("joining queue thread")
        thread = twr()
        if thread is not None:
            thread.join()
            debug("... queue thread joined")
        else:
            debug("... queue thread already dead")

    @staticmethod
    def _finalize_close(buffer, notempty):
        debug("telling queue thread to quit")
        with notempty:
            buffer.append(_sentinel)
            notempty.notify()

    @staticmethod
    def _feed(
        buffer,
        notempty,
        send_bytes,
        writelock,
        close,
        ignore_epipe,
        onerror,
        queue_sem,
        qsize,
    ):
        debug("starting thread to feed data to pipe")
        nacquire = notempty.acquire
        nrelease = notempty.release
        nwait = notempty.wait
        bpopleft = buffer.popleft
        sentinel = _sentinel
        if sys.platform != "win32":
            wacquire = writelock.acquire
            wrelease = writelock.release
        else:
            wacquire = None

        while 1:
            try:
                nacquire()
                try:
                    if not buffer:
                        nwait()
                finally:
                    nrelease()
                try:
                    while 1:
                        obj = bpopleft()
                        if obj is sentinel:
                            debug("feeder thread got sentinel -- exiting")
                            close()
                            return

                        # serialize the data before acquiring the lock
                        obj = _ForkingPickler.dumps(obj)
                        if wacquire is None:
                            send_bytes(obj)
                        else:
                            wacquire()
                            try:
                                send_bytes(obj)
                            finally:
                                wrelease()
                except IndexError:
                    pass
            except Exception as e:
                if ignore_epipe and getattr(e, "errno", 0) == errno.EPIPE:
                    return
                # Since this runs in a daemon thread the resources it uses
                # may be become unusable while the process is cleaning up.
                # We ignore errors which happen after the process has
                # started to cleanup.
                if is_exiting():
                    info("error in queue thread: %s", e)
                    return
                else:
                    # Since the object has not been sent in the queue, we need
                    # to decrease the size of the queue. The error acts as
                    # if the object had been silently removed from the queue
                    # and this step is necessary to have a properly working
                    # queue.
                    queue_sem.release()
                    qsize.value -= 1
                    onerror(e, obj)

    @staticmethod
    def _on_queue_feeder_error(e, obj):
        """
        Private API hook called when feeding data in the background thread
        raises an exception.  For overriding by concurrent.futures.
        """
        import traceback

        traceback.print_exc()



import atexit
import time
import typing as t
from weakref import WeakSet

from hybrid_pool_executor.base import (
    BaseExecutor,
    BaseManager,
    BaseManagerSpec,
    Future,
    ModuleSpec,
    NotSupportedError,
)
from hybrid_pool_executor.constants import Function
from hybrid_pool_executor.spec import ModuleSpecRepo, spec_factory
from hybrid_pool_executor.utils import isasync, iscoroutine

_all_executors: WeakSet = WeakSet()


def _python_exit():
    for executor in _all_executors:
        if executor.is_alive():
            executor.shutdown()


atexit.register(_python_exit)


class HybridPoolExecutor(BaseExecutor):
    def __init__(
        self,
        thread_workers: int = -1,
        incremental_thread_workers: bool = True,
        thread_worker_name_pattern: t.Optional[str] = None,
        redirect_thread: t.Optional[str] = None,
        process_workers: int = -1,
        incremental_process_workers: bool = True,
        process_worker_name_pattern: t.Optional[str] = None,
        redirect_process: t.Optional[str] = None,
        async_workers: int = -1,
        incremental_async_workers: bool = True,
        async_worker_name_pattern: t.Optional[str] = None,
        redirect_async: t.Optional[str] = None,
        extra_specs: t.Optional[t.Iterable[ModuleSpec]] = None,
        **kwargs,
    ):
        self._spec_repo: ModuleSpecRepo = spec_factory.get_repo()
        if extra_specs is not None:
            for spec in extra_specs:
                self._spec_repo.import_spec(spec, overwrite=True)
        self._managers: t.Dict[str, BaseManager] = {}
        self._manager_kwargs = {
            "thread_workers": thread_workers,
            "incremental_thread_workers": incremental_thread_workers,
            "thread_worker_name_pattern": thread_worker_name_pattern,
            "redirect_thread": redirect_thread,
            "process_workers": process_workers,
            "incremental_process_workers": incremental_process_workers,
            "process_worker_name_pattern": process_worker_name_pattern,
            "redirect_process": redirect_process,
            "async_workers": async_workers,
            "incremental_async_workers": incremental_async_workers,
            "async_worker_name_pattern": async_worker_name_pattern,
            "redirect_async": redirect_async,
            **kwargs,
        }

        self._last_clear_ts: float = time.monotonic()
        self._clear_counter: int = 0

        global _all_executors
        _all_executors.add(self)
        self._is_alive: bool = True

    @property
    def specs(self) -> ModuleSpecRepo:
        return self._spec_repo

    @classmethod
    def _get_manager(
        cls,
        mode: str,
        module_spec: ModuleSpec,
        kwargs: t.Dict[str, t.Any],
    ) -> BaseManager:
        if (redirect := kwargs.get(f"redirect_{mode}")) is not None:
            mode = redirect
        manager_spec: BaseManagerSpec = module_spec.manager_spec_type()
        if (num_workers := kwargs.get(f"{mode}_workers")) is not None:
            manager_spec.num_workers = num_workers
        if (incremental := kwargs.get(f"incremental_{mode}_workers")) is not None:
            manager_spec.incremental = incremental
        if (
            worker_name_pattern := kwargs.get(f"{mode}_worker_name_pattern")
        ) is not None:
            manager_spec.worker_name_pattern = worker_name_pattern
        return module_spec.manager_type(manager_spec)

    def submit(  # type: ignore
        self,
        fn: t.Callable[..., t.Any],
        /,
        *args,
        **kwargs,
    ) -> Future:
        return self.submit_task(
            fn,
            args=args,
            kwargs=kwargs,
            name=kwargs.get("_name"),
            mode=kwargs.get("_mode"),
            tags=kwargs.get("_tags"),
        )

    def apply_async(
        self,
        func: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> Future:
        return self.submit(func, *args, **(kwargs or {}))

    def apply(
        self,
        func: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Any:
        return self.submit(func, *args, **(kwargs or {})).result()

    def map_tasks(
        self,
        fn: t.Callable[..., t.Any],
        *iterables: t.Union[t.Iterable[t.Any], t.Mapping[str, t.Any], t.Any],
        timeout: t.Optional[float] = None,
        mode: t.Optional[str] = None,
        tags: t.Optional[t.Iterable[str]] = None,
    ) -> t.Generator[t.Any, None, None]:
        if timeout is not None:
            end_time = timeout + time.monotonic()
        fs = []
        for params in iterables:
            if isinstance(params, t.Mapping):
                fs.append(
                    self.submit_task(
                        fn=fn,
                        kwargs=t.cast(t.Dict[str, t.Any], params),
                        name=params.get("_name"),
                        mode=params.get("_mode", mode),
                        tags=params.get("_tags", tags),
                    )
                )
            else:
                if not isinstance(params, t.Iterable):
                    params = [params]
                fs.append(self.submit_task(fn=fn, args=params, mode=mode, tags=tags))

        def result_iterator() -> t.Any:
            try:
                fs.reverse()
                while fs:
                    if timeout is None:
                        yield fs.pop().result()
                    else:
                        yield fs.pop().result(end_time - time.monotonic())
            finally:
                for future in fs:
                    future.cancel()

        return result_iterator()

    @t.overload
    def submit_task(
        self,
        fn: t.Coroutine[t.Any, t.Any, t.Any],
        *,
        name: t.Optional[str] = None,
        mode: t.Optional[str] = None,
        tags: t.Optional[t.Iterable[str]] = None,
    ) -> Future:
        ...

    @t.overload
    def submit_task(
        self,
        fn: t.Callable[..., t.Any],
        args: t.Optional[t.Iterable[t.Any]] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
        mode: t.Optional[str] = None,
        tags: t.Optional[t.Iterable[str]] = None,
    ) -> Future:
        ...

    def submit_task(
        self,
        fn: Function,
        args: t.Optional[t.Iterable[t.Any]] = (),
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        name: t.Optional[str] = None,
        mode: t.Optional[str] = None,
        tags: t.Optional[t.Iterable[str]] = None,
        **_,
    ) -> Future:
        if iscoroutine(fn) and (args or kwargs):
            raise ValueError("Coroutine should not have arguments specified.")
        modes = self._guess_mode(fn, mode, tags)
        if not modes:
            raise NotSupportedError(
                (
                    f"No match mode found for task {fn}{name} with requirements: "
                    f'["mode={mode}, tags={tags}"].'
                ).format(
                    fn=fn,
                    name=f' "{name}"' if name else "",
                    mode=mode,
                    tags=tags,
                )
            )
        mode = modes.__iter__().__next__()
        manager = self._managers.get(mode)
        if not manager or not manager.is_alive():
            module_spec: ModuleSpec = self._spec_repo[mode]
            manager = self._get_manager(
                mode=mode,
                module_spec=module_spec,
                kwargs=self._manager_kwargs,
            )
            self._managers[mode] = manager
        self._is_alive = True
        if not manager.is_alive():
            manager.start()
        self._clear_manager_if_needed()
        return (
            manager.submit(
                fn=t.cast(t.Coroutine[t.Any, t.Any, t.Any], fn),
                name=name,
            )
            if iscoroutine(fn)
            else manager.submit(
                fn=t.cast(t.Callable[..., t.Any], fn),
                args=args,
                kwargs=kwargs,
                name=name,
            )
        )

    def _clear_manager_if_needed(self) -> None:
        self._clear_counter += 1
        monotonic = time.monotonic()
        now, prev, self._last_clear_ts = monotonic, self._last_clear_ts, monotonic
        if self._clear_counter <= 256 and now - prev < 300:
            return
        self._clear_counter = 0
        cleared = []
        for name, manager in self._managers.items():
            if not manager.is_alive():
                cleared.append(name)
        for name in cleared:
            self._managers.pop(name)

    def _guess_mode(
        self,
        fn: Function,
        mode: t.Optional[str] = None,
        tags: t.Optional[t.Iterable[str]] = None,
    ) -> t.FrozenSet[str]:
        if tags:
            mode_filter = self._spec_repo.filter_by_tags(*tags)
        else:
            mode_filter = None
        if mode:
            if mode_filter is not None:
                res = {mode} if mode in mode_filter else frozenset()
            else:
                res = {mode}
        else:
            if mode_filter is not None:
                res = mode_filter
            else:
                res = {"async"} if isasync(fn) else {"thread"}
        if res == {"async"} and not isasync(fn):
            return frozenset()
        return frozenset(res)

    def is_alive(self) -> bool:
        return self._is_alive

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False):
        for manager in self._managers.values():
            if wait:
                manager.stop()
            else:
                manager.terminate()
        self._is_alive = False



import asyncio
import sys
import time
from random import random

import pytest

from hybrid_pool_executor import HybridPoolExecutor


def simple_task():
    return "done"


def simple_delay_task(v):
    time.sleep(random())
    return v


async def simple_async_delay_task(v):
    await asyncio.sleep(random())
    return v


@pytest.mark.timeout(10)
def test_executor_simple():
    pool = HybridPoolExecutor()
    future = pool.submit(simple_task)
    assert future.result() == "done"


@pytest.mark.timeout(10)
def test_executor_map():
    pool = HybridPoolExecutor()
    results = pool.map(simple_delay_task, range(4))
    for i, result in enumerate(results):
        assert result == i


@pytest.mark.timeout(10)
def test_executor_map_tasks():
    pool = HybridPoolExecutor()
    results = pool.map(simple_delay_task, [(i, i) for i in range(4)])
    for i, result in enumerate(results):
        assert result == (i, i)


def test_executor_guess_mode():
    pool = HybridPoolExecutor()
    guess_mode = pool._guess_mode

    assert "thread" in guess_mode(simple_delay_task)
    assert "async" in guess_mode(simple_async_delay_task)
    assert guess_mode(simple_async_delay_task, mode="async") == {"async"}
    assert guess_mode(simple_delay_task, mode="thread") == {"thread"}
    assert guess_mode(simple_delay_task, mode="process") == {"process"}
    assert len(guess_mode(simple_delay_task, mode="async")) == 0

    assert "thread" in guess_mode(simple_async_delay_task, tags=["thread"])
    assert "process" in guess_mode(simple_async_delay_task, tags=["thread"])
    assert "thread" in guess_mode(simple_delay_task, tags=["thread"])
    assert "process" in guess_mode(simple_delay_task, tags=["thread"])
    assert "thread" not in guess_mode(simple_delay_task, tags=["process"])
    assert "process" in guess_mode(simple_delay_task, tags=["process"])

    assert len(guess_mode(simple_delay_task, tags=["black hole"])) == 0


@pytest.mark.skipif(sys.platform == "darwin", reason="test unstable on darwin")
@pytest.mark.timeout(20 if sys.platform == "linux" else 60)
@pytest.mark.asyncio
async def test_executor_high_concurrency():
    futures = {
        "thread": [],
        "process": [],
        "async": [],
    }
    total_tasks, process_tasks = (64, 16) if sys.platform == "linux" else (16, 4)
    # NOTE: The more threads are created, the more time they takes to stop, and the
    # longer the pytest timeout tolerance is required.
    with HybridPoolExecutor() as pool:
        for i in range(total_tasks):
            if i < process_tasks:
                futures["process"].append(
                    pool.submit_task(fn=simple_delay_task, args=(i,), mode="process")
                )
            futures["thread"].append(
                pool.submit_task(fn=simple_delay_task, args=(i,), mode="thread")
            )
            futures["async"].append(
                pool.submit_task(fn=simple_async_delay_task, args=(i,), mode="async")
            )
            # TODO: process futures need syncing among processes by manager, code will
            #       be probably blocked here to wait for all process futures to be set.
    for i in range(total_tasks):
        if i < process_tasks:
            assert await futures["process"][i] == i
        assert await futures["thread"][i] == i
        assert await futures["async"][i] == i


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_executor_run_in_executor():
    pool = HybridPoolExecutor()
    loop = asyncio.get_event_loop()
    assert await loop.run_in_executor(pool, simple_task) == "done"



import typing as t
from multiprocessing.queues import Queue as ProcessQueue
from queue import Queue, SimpleQueue

Function = t.Union[t.Callable[..., t.Any], t.Coroutine[t.Any, t.Any, t.Any]]
ThreadBus = t.Union[Queue, SimpleQueue]
ThreadBusType = t.Union[t.Type[Queue], t.Type[SimpleQueue], t.Callable[[], ThreadBus]]
ProcessBus = ProcessQueue
ProcessBusType = t.Union[t.Type[ProcessBus], t.Callable[[], ProcessBus]]

PRESERVED_TASK_TAGS = frozenset({"async", "process", "thread"})

ActionFlag = int
ACT_NONE = 0
ACT_DONE = 1
ACT_CLOSE = 1 << 1
ACT_EXCEPTION = 1 << 2
ACT_RESTART = 1 << 3
ACT_RESET = 1 << 4
ACT_TIMEOUT = 1 << 5
ACT_CANCEL = 1 << 6
ACT_COERCE = 1 << 7
